/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999-2004 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *
 * --------------------------------------------------------------------------
 * $Id: Server.java,v 1.96 2004/09/28 16:33:53 durieuxp Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.server;

import java.rmi.RMISecurityManager;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Properties;
import java.util.StringTokenizer;

import javax.management.JMException;
import javax.management.MBeanServer;
import javax.management.ObjectName;
import javax.management.modelmbean.ModelMBean;
import javax.naming.InitialContext;
import javax.security.jacc.PolicyContext;
import javax.security.jacc.PolicyContextException;
import javax.security.jacc.PolicyContextHandler;

import org.apache.commons.modeler.ManagedBean;
import org.apache.commons.modeler.Registry;
import org.objectweb.carol.util.configuration.CarolConfiguration;
import org.objectweb.carol.util.configuration.CarolCurrentConfiguration;
import org.objectweb.jonas.adm.Adm;
import org.objectweb.jonas.common.JProp;
import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.common.LogManagement;
import org.objectweb.jonas.jmx.J2eeObjectName;
import org.objectweb.jonas.jmx.JmxService;
import org.objectweb.jonas.jmx.JonasObjectName;
import org.objectweb.jonas.management.JonasMBeanTools;
import org.objectweb.jonas.management.ReconfigManager;
import org.objectweb.jonas.management.j2eemanagement.J2EEDomain;
import org.objectweb.jonas.naming.JNDIResource;
import org.objectweb.jonas.naming.NamingManager;
import org.objectweb.jonas.security.PolicyProvider;
import org.objectweb.jonas.security.jacc.JPolicyContextHandler;
import org.objectweb.jonas.service.JonasAlreadyStartedException;
import org.objectweb.jonas.service.ServiceException;
import org.objectweb.jonas.service.ServiceManager;
import org.objectweb.jonas_lib.I18n;
import org.objectweb.jonas_lib.deployment.work.WorkCleaner;
import org.objectweb.jonas_lib.version.Version;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.wrapper.javaLog.MonologFormatter;

/**
 * This class implements an JOnAS Server.
 *
 * @author Philippe Coq, Philippe Durieux 00/06/15: Jeff Mesnil 01/06/15: Regis
 *         Le Brettevillois - Libelis / JOnAS team - evidian Service as
 *         org.objectweb.jonas.service.Service 02/04/25: Sebastien
 *         Chassande-Barrioz sebastien.chassande@inrialpes.fr Monolog for traces
 *
 * Adriana Danes 03/01/20: -Highlight configuration properties 03/03/..: -Force
 * jmx service startup -Create and register the JonAS configuration MBean. This
 * MBean manages JOnAS persistent re-configuration (see ReconfigManager class)
 */
public class Server {

    /**
     * Current thread group
     */
    ThreadGroup thGroup = null;

    /**
     * Property value for the default file
     */
    private static final String DEF_LOG_CONFIGFILE = "trace";

    /**
     * Carol default ORB port
     */
    private static final String DEF_ORB_PORT = "0";

    /**
     * Internationalization
     */
    private static I18n i18n = I18n.getInstance(Server.class);

    /**
     * Property for the configuration log file
     */
    private static final String LOG_CONFIGFILE = "jonas.log.configfile";

    /**
     * Logger
     */
    private static Logger logger = null;

    /**
     * Server configuration properties
     */
    private static final String ORB_PORT = "jonas.orb.port";

    /**
     * Property for the security propagation
     */
    private static final String SECURITY_PROPAGATION = "jonas.security.propagation";

    /**
     * Property for the security manager
     */
    private static final String SECURITY_MANAGER = "jonas.security.manager";
    
    /**
     * J2EE Application Server Name
     */
    private static final String SERVERNAME = "JOnAS";

    /**
     * Default sleep value
     */
    private static final int SLEEP_VALUE = 10000;

    /**
     * Property for the transaction propagation
     */
    private static final String TRANSACTION_PROPAGATION = "jonas.transaction.propagation";

    /**
     * The Server singleton value
     */
    private static Server unique = null;

    /**
     * Get the unique instance. Create it at first call.
     * @return unique instance.
     * @throws Exception if the Server can not be instanciate
     */
    public static Server getInstance() throws Exception {
        if (unique == null) {
            unique = new Server();
        }
        return unique;
    }

    /**
     * Server main routine. No args.
     * @param args the arguments of the server
     * @exception Exception can not start server
     */
    public static void main(String[] args) throws Exception {
        Server server = null;

        // Add Mx4J logger as wrapper (allow to put the right class for debug in the log)
        System.setProperty(MonologFormatter.WRAPPERS_PROPERTY, "mx4j.log.CommonsLogger,mx4j.log.Logger");

        try {
            server = Server.getInstance();
            server.start();
        } catch (JonasAlreadyStartedException e) {
            // if we get this exception this means that
            // we are trying to restart the same JOnAS
            System.err.println(i18n.getMessage("Server.main.alreadyStarted", SERVERNAME));
            // rethrow JonasAlreadyStartedException
            throw e;
        } catch (Exception e) {
            // We don't use here the trace system.
            System.err.println("JOnAS error: " + e);
            e.printStackTrace(System.err);
            System.err.println(SERVERNAME + " halting");
            try {
                if (server != null) {
                    server.stop();
                }
            } catch (Exception he) {
                System.err.println("Exception when halting" + he.getMessage());
                throw he;
            }
            throw e;
        }
    }

    /**
     * Associated Adm object
     */
    private Adm adm = null;

    /**
     * J2EE Management domain
     */
    private String domainName = null;


    /**
     * J2EEServer associate
     */
    private J2EEServer j2EEServer = null;

    /**
     * File for log configuration
     */
    private String logConfigFile;

    /**
     * JOnAS configuration properties
     */
    private JProp props = null;

    /**
     * Current server name (the server instance name)
     */
    private String serverName = null;

    /**
     * The associated ServiceManager for JOnAS
     */
    private ServiceManager sm = null;

    /**
     * Used to keep naming context properties for
     * the different protocols used by Carol
     */
    private HashMap environments = new HashMap();

    /**
     * Number of protocols declared in Carol configuration.
     * If > 1, JOnAS works in multi-protocol
     */
    private int nbCarolProtocols = 0;

    /**
     * Private constructor to force only 1 instance.
     *
     * @throws Exception (for example if JProp can't read properties file)
     */
    private Server() throws Exception {
        // read the JOnAS properties
        props = JProp.getInstance();

        // set configuration info into the Log System (monolog)
        logConfigFile = props.getValue(LOG_CONFIGFILE, DEF_LOG_CONFIGFILE);
        Log.configure(logConfigFile);

        // get a logger for server traces
        logger = Log.getLogger(Log.JONAS_SERVER_PREFIX);

        // create the service manager
        sm = ServiceManager.getInstance();

        // called only once for a JVM
        PolicyProvider.init();

    }

    /**
     * @return the estimated number of active groups in the JOnAS server's initial
     * thread group
     */
    public int getActiveGroupCount() {
        return thGroup.activeGroupCount();
    }

    /**
     * @return the number of active threads in the JOnAS server's initial thread
     * group
     */
    protected int getActiveThreadsCount() {
        int ret = Thread.activeCount() - 2;
        return ret > 1 ? ret : 1;
    }

    /**
     * Gets J2EEServer.
     *
     * @return J2EEServer
     */
    public J2EEServer getJ2EEServer() {
        return j2EEServer;
    }

    /**
     * Initialize interception and carol
     *
     * @param props the JOnAS properties
     * @return list of protocols
     * @throws Exception if carol init failed
     */
    public String initCarol(JProp props) throws Exception {

        // Intialize Carol
        CarolConfiguration.init(Log.getLoggerFactory());

        // Initialize Interceptors
        String sec = props.getValue(SECURITY_PROPAGATION, "false").trim();
        String trans = props.getValue(TRANSACTION_PROPAGATION, "false").trim();
        if (sec.equals("true")) {
            // jrmp interceptors
            CarolConfiguration.addInterceptors("jrmp",
                    "org.objectweb.jonas.security.interceptors.jrmp.SecurityInitializer");
            // cmi interceptors
            CarolConfiguration.addInterceptors("cmi",
                    "org.objectweb.jonas.security.interceptors.jrmp.SecurityInitializer");
            // iiop interceptors flag
            CarolConfiguration.addInterceptors("iiop",
                    "org.objectweb.jonas.security.interceptors.iiop.SecurityInitializer");

        }

        if (trans.equals("true")) {
            // jrmp interceptors flag
            CarolConfiguration.addInterceptors("jrmp", "org.objectweb.jotm.jta.rmi.JTAInterceptorInitializer");
            // jrmp interceptors flag
            CarolConfiguration.addInterceptors("cmi", "org.objectweb.jotm.jta.rmi.JTAInterceptorInitializer");
            // iiop interceptors flag
            CarolConfiguration.addInterceptors("iiop", "org.objectweb.jotm.ots.OTSORBInitializer");
        }

        // intialize port number
        System.setProperty("rmi.server.port", props.getValue(ORB_PORT, DEF_ORB_PORT));

        // get Carol configuration
        CarolCurrentConfiguration carolConfig = CarolCurrentConfiguration.getCurrent();
        String s = CarolConfiguration.getProtocols();
        StringTokenizer st = new StringTokenizer(s, ",", false);
        String pn = " ";

        int idx = 0;
        while (st.hasMoreElements()) {
            String token = st.nextToken().trim();
            if (idx != 0) {
                pn = pn + ", ";
            }
            pn = pn + "rmi/" + token;
            idx++;
            // initialize environments by creating one environment for each protocol
            Properties protocolProps = carolConfig.getRMIProperties(token);
            if (protocolProps == null) {
                throw new Exception("no protocol properties for " + token);
            }
            Properties environment = new Properties();
            for (Enumeration pNames = protocolProps.propertyNames(); pNames.hasMoreElements();) {
                String pName = (String) pNames.nextElement();
                String pValue = protocolProps.getProperty(pName);
                if ((pName != null) && (pValue != null)) {
                    environment.setProperty(pName, pValue);
                }
            }
            environments.put(token, environment);
        }
        nbCarolProtocols = idx;
        return pn;
    }

    /**
     * Initialize the context handlers (JACC)
     *
     * @throws PolicyContextException if handlers can't be registered
     */
    private void initJACCPolicyContextHandlers() throws PolicyContextException {
        PolicyContextHandler policyContextHandler = new JPolicyContextHandler();
        String[] keys = policyContextHandler.getKeys();
        for (int k = 0; k < keys.length; k++) {
            PolicyContext.registerHandler(keys[k], policyContextHandler, true);
        }
    }

    /**
     * Stop the server and stop the JVM (MBean method)
     *
     * @throws Exception if the server can't be destroyed
     */
    public void kill() throws Exception {
        if (adm != null) {
            adm.killServer();
        }
    }

    /**
     * Start the server
     *
     * @throws Exception if server can't be started
     */
    public void start() throws Exception {

        // Server name
        serverName = props.getValue(JProp.JONAS_NAME, JProp.JONAS_DEF_NAME);

        // Initialize Context Interceptors and Carol
        String protocols = initCarol(props);

        // RMI Security Manager
        boolean useSecurityManager = new Boolean(props.getValue(SECURITY_MANAGER, "true").trim()).booleanValue();

        if (useSecurityManager) {
            if (System.getSecurityManager() == null) {
                System.setSecurityManager(new RMISecurityManager());
            }
        } else {
            logger.log(BasicLevel.INFO, "Security manager is not set.");
        }

        // Initialize Policy Context handlers (JACC)
        try {
            initJACCPolicyContextHandlers();
        } catch (PolicyContextException pce) {
            logger.log(BasicLevel.ERROR, "Could not register JOnAS Policy Context Handlers");
            throw pce;
        }

        // First of all, start registry service
        sm.startRegistry();

        // then, start JMX service (start an MBean server and create RMI
        // connector for it registered in JNDI)
        sm.startJmx();

        // start the cleaner thread (deleting unused working files)
        //get the cleaner ref
        WorkCleaner.getInstance().start();

        // get the JMX server reference
        JmxService jmxService = null;
        MBeanServer mbeanServer = null;
        try {
            jmxService = (JmxService) ServiceManager.getInstance().getJmxService();
            mbeanServer = jmxService.getJmxServer();
        } catch (Exception se) {
            // this could mean that jmx service not available
            // this is not possible, as we started it even if it does not apper
            // in the services list configuration parameter
            logger.log(BasicLevel.ERROR, "Could not access the jmx service");
            throw se;
        }

        // -----------------------------------------------------
        // J2EE Management
        // -----------------------------------------------------

        // Get the MBean register (and load mbeans-descriptors)
        Registry oRegistry = JonasMBeanTools.getRegistry();

        // Create and register a MBean for the Management Domain
        domainName = props.getValue(JProp.DOMAIN_NAME);

        ObjectName domainOn = null;
        boolean domainOK = false;
        try {
            // Create a JMX object name for the Domain MBean
            domainOn = J2eeObjectName.J2EEDomain(domainName);
            J2EEDomain j2eeDomain = new J2EEDomain(domainOn.toString(), false, false, false);
            ManagedBean oManaged = oRegistry.findManagedBean("J2EEDomain");
            ModelMBean oMBean = oManaged.createMBean(j2eeDomain);
            mbeanServer.registerMBean(oMBean, domainOn);
            domainOK = true;
        } catch (javax.management.InstanceAlreadyExistsException ae) {
            // should never occur
            logger.log(BasicLevel.DEBUG, ae.getMessage());
        } catch (Exception e) {
            logger.log(BasicLevel.WARN, "Could not register J2EEDomain MBean");
            domainOK = false;
        }

        // create and register necessary MBeans for the server management
        ObjectName serverOn = null;
        try {
            // create and register a MBean allowing server reconfiguration
            // management
            ReconfigManager reconfigManagerMBean = new ReconfigManager(props.getConfigFileEnv(), mbeanServer);
            mbeanServer.registerMBean(reconfigManagerMBean, JonasObjectName.serverConfig());

            // create and register a MBean allowing the JOnAS server
            // administration
            serverOn = J2eeObjectName.J2EEServer(domainName, serverName);
            j2EEServer = new J2EEServer(serverOn.toString(), this, serverName, Version.NUMBER, Version.VENDOR,
                    protocols);

            // Add j2EEServer object as listner to  MBeanServerNotifications.
            // Use for this the MBeanServerDelegate ObjectName as
            // argument in the addNotificationListener method (see JMX API for more info).
            // Use null NotificationFilter and null handback object.
            try {
                ObjectName delegate = new ObjectName("JMImplementation:type=MBeanServerDelegate");
                mbeanServer.addNotificationListener(delegate, j2EEServer, null, null);
            } catch (JMException me) {
                // MalformedObjectNameException should not occur if the JMX implementation is correct
                // InstanceNotFoundException should not occur as we use MBeanServerDelegate ObjectName
                // TODO
            }

            ManagedBean oManaged = oRegistry.findManagedBean("J2EEServer");
            ModelMBean oMBean = oManaged.createMBean(j2EEServer);
            mbeanServer.registerMBean(oMBean, serverOn);

            // create and register a MBean for the logging service
            // administration
            LogManagement logMBean = LogManagement.getInstance();
            mbeanServer.registerMBean(logMBean, JonasObjectName.logService(logConfigFile));
        } catch (Exception e) {
            // Cannot administrate this server through JMX
            jmxService.stop();
        }

        // Update J2EEDomain MBean
        // ------------------------
        if (domainOK) {
            try {
                String operationName = "addServer";
                Object[] params = {serverOn.toString()};
                String[] signature = {"java.lang.String"};
                // use the domainOn ObjectName
                // this code is related to the temporary implementation as said
                // before
                mbeanServer.invoke(domainOn, operationName, params, signature);
            } catch (Exception e) {
                logger.log(BasicLevel.WARN, "Could not update J2EEDomain MBean with J2EE server object name "
                        + serverOn.toString());
            }
        }

        // Create the MBean for the corresponding JVM
        ObjectName onJvm = null;
        try {
            onJvm = J2eeObjectName.JVM(domainName, serverName, serverName);
            ManagedBean oManaged = oRegistry.findManagedBean("JVM");
            JavaVm oJvm = new JavaVm(onJvm.toString(), props);
            ModelMBean oMBean = oManaged.createMBean(oJvm);
            mbeanServer.registerMBean(oMBean, onJvm);
            // Update J2EEServer MBean
            j2EEServer.addJavaVM(onJvm.toString());
        } catch (Exception e) {
            logger.log(BasicLevel.WARN, "Could not create JVM MBean");
        }

        ObjectName on = null;
        // create and register JNDI resource MBeans
        ManagedBean oManaged = oRegistry.findManagedBean("JNDIResource");
        if (nbCarolProtocols == 1) {
            // keep initial solution (use NamingManager)
            try {
                on = J2eeObjectName.JNDIResource(domainName, "jndi", serverName);
                // JNDI resource creation + registration
                NamingManager naming = NamingManager.getInstance();
                JNDIResource resMBean = new JNDIResource(on.toString(), naming.getInitialContext());
                ModelMBean oMBean = oManaged.createMBean(resMBean);
                mbeanServer.registerMBean(oMBean, on);
            } catch (Exception e) {
                logger.log(BasicLevel.WARN, "Problem with J2EEResource MBean " + on.toString()
                        + " creation or registration: " + e);
            }
        } else {
            // Treat multi-protocol case
            InitialContext protocolContext = null;
            for (Iterator it = environments.keySet().iterator(); it.hasNext();) {
                String protocol = (String) it.next();
                on = J2eeObjectName.JNDIResource(domainName, "jndi_" + protocol, serverName);
                protocolContext = new InitialContext((Properties) environments.get(protocol));
                JNDIResource resMBean = new JNDIResource(on.toString(), protocolContext);
                ModelMBean oMBean = oManaged.createMBean(resMBean);
                mbeanServer.registerMBean(oMBean, on);
            }
        }

        // Creates the Adm instance for remote administration
        try {
            adm = new Adm(props);
        } catch (Exception e) {
            e.printStackTrace();
            throw new Exception("Cannot create Adm : " + e);
        }

        // start all other services
        sm.startServices();

        // TM Server or EJB Server ?
        boolean isEJB = true;
        try {
            sm.getEjbService();
        } catch (ServiceException se) {
            isEJB = false;
        }
        adm.serverReady(isEJB);
        System.out.println(i18n.getMessage("Server.start.serverIsReady", serverName, Version.NUMBER));
        logger.log(BasicLevel.INFO, i18n.getMessage("Server.start.serverStarted", serverName, protocols));

        thGroup = Thread.currentThread().getThreadGroup();

        // Don't stop this thread to keep the server alive.
        // Could do here some statistics for JMX.
        while (true) {
            Thread.sleep(SLEEP_VALUE);
            //printThreadList(thGroup);
        }
    }

    /**
     * Get the list of all threads in thsi thread group.
     * @param tg thread group
     */
    public void printThreadList(ThreadGroup tg) {
        logger.log(BasicLevel.DEBUG, "thread count=" + Thread.activeCount());

        // Get the number of active threads
        int count = tg.activeCount();

        // Create an array of threads, to hold all of the active threads
        Thread[] array = new Thread[count];

        // Copy to array
        tg.enumerate(array);

        for (int i = 0; i < count; i++) {
            if (array[i] != null) {
                logger.log(BasicLevel.DEBUG, "Thread " + i + ": " + array[i].getName());
            }
        }
    }

    /**
     * Stop the server without stopping the JVM
     *
     * @throws Exception if the server can't be stopped
     */
    public void stop() throws Exception {
        if (adm != null) {
            System.out.println(i18n.getMessage("Server.stop.serverIsStopped", serverName));
            logger.log(BasicLevel.INFO, i18n.getMessage("Server.stop.serverStopped", serverName));
            adm.stopServer();
        }
    }

}
