/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999-2004 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: JAxisServiceFactory.java,v 1.14 2004/09/10 11:37:56 sauthieg Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.ws.axis;

import java.io.IOException;
import java.io.InputStream;
import java.io.StringWriter;
import java.lang.reflect.Constructor;
import java.lang.reflect.Proxy;
import java.net.URL;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.Name;
import javax.naming.NamingException;
import javax.naming.Reference;
import javax.naming.StringRefAddr;
import javax.wsdl.Port;
import javax.wsdl.extensions.ExtensibilityElement;
import javax.wsdl.extensions.soap.SOAPAddress;
import javax.xml.namespace.QName;

import org.w3c.dom.Document;

import org.apache.axis.ConfigurationException;
import org.apache.axis.EngineConfiguration;
import org.apache.axis.client.AxisClient;
import org.apache.axis.configuration.XMLStringProvider;
import org.apache.axis.deployment.wsdd.WSDDDeployment;
import org.apache.axis.encoding.SerializationContext;
import org.apache.axis.utils.XMLUtils;
import org.apache.axis.wsdl.gen.Parser;
import org.apache.axis.wsdl.symbolTable.ServiceEntry;
import org.apache.axis.wsdl.toJava.Utils;

import org.objectweb.jonas_lib.I18n;

import org.objectweb.jonas_ws.deployment.api.MappingFile;
import org.objectweb.jonas_ws.deployment.api.PortComponentDesc;
import org.objectweb.jonas_ws.deployment.api.PortComponentRefDesc;
import org.objectweb.jonas_ws.deployment.api.ServiceRefDesc;

import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.ws.JServiceFactory;
import org.objectweb.jonas.ws.WSServiceException;

import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * Axis specific JServiceFactory. Create an Axis Service Reference.
 * Used to recreate Axis Service bound in JNDI (ObjectFactory)
 *
 * @author Guillaume Sauthier
 * @author Xavier Delplanque
 */
public class JAxisServiceFactory implements JServiceFactory {

    /** logger */
    private static Logger logger = Log.getLogger(Log.JONAS_WS_PREFIX);

    /** i18n instance */
    private static I18n i18n = I18n.getInstance(JAxisServiceFactory.class);

    /** The jonas-init-param name for client configuration file */
    private static final String AXIS_CLIENT_CONFIG_PARAM = "axis.clientConfigFile";

    /** default client configuration file to be filled */
    private static final String CLIENT_CONFIG_WSDD =
        "org/objectweb/jonas/ws/axis/client-config.wsdd";

    /** RefAddr name for configuration storing */
    private static final String REF_WSDD = "client-config.wsdd";

    /** RefAddr name for wsdl storing */
    private static final String REF_WSDL = "wsdl-filename";

    /** RefAddr name for wsdl URL storing */
    private static final String REF_WSDL_URL = "wsdl-url";

    /** RefAddr name for Service QName */
    private static final String REF_SERVICE_QNAME_NS =
        "service-qname-namespace";

    /** RefAddr name for Service QName */
    private static final String REF_SERVICE_QNAME_LP =
        "service-qname-localpart";

    /** RefAddr name for endpoint URL */
    private static final String REF_URL_PREFIX = "URL-";

    /** RefAddr name port link */
    private static final String REF_PORT_LINK_LIST = "port-link-list";

    /** Service Implementation in JOnAS */
    private static final String JONAS_SERVICE_CLASSNAME =
        "org.objectweb.jonas.ws.axis.JService";

    /** List separator */
    private static final String LIST_SEP = ",";

    /**
     * WSDL Directory in Jar files
     */
    private static final String EJB_WSDL_DIR = "META-INF/wsdl/";

    /**
     * WSDL Directory in WebApp files
     */
    private static final String WEB_WSDL_DIR = "WEB-INF/wsdl/";

    /**
     * Construct a new JAxisServiceFactory.
     */
    public JAxisServiceFactory() { }

    /**
     * Create a Reference on a Service from an Axis Service object.
     *
     * @param sr the service-ref instance where informations are extracted
     * @param cl the classloader where config files can be loaded
     *
     * @return a Reference on a Service from an Axis Service object.
     *
     * @throws WSServiceException if wsdd cannot be load or merged
     */
    public Reference getServiceReference(ServiceRefDesc sr,
                                         ClassLoader cl)
        throws WSServiceException {

        String classname = createServiceClassname(sr);

        logger.log(BasicLevel.DEBUG, "Service class: '" + classname + "'");

        // Create a reference on the Service class (generated or not)
        Reference ref = new Reference(classname,
                                      getClass().getName(),
                                      null);

        WSDDDeployment base = loadAxisDeployment(CLIENT_CONFIG_WSDD, cl);

        // get the WSDD configuration
        String clientConfig = sr.getParam(AXIS_CLIENT_CONFIG_PARAM);

        // If ServiceRef has a specified ConfigFile
        if (clientConfig != null) {
            // load wsdd
            WSDDDeployment deploy = loadAxisDeployment(clientConfig, cl);

            // merge with base
            try {
                deploy.deployToRegistry(base);
            } catch (ConfigurationException wsdde) {
                String err = i18n.getMessage("JAxisServiceFactory.getServiceReference.mergeFail",
                                             sr.getServiceRefName());
                logger.log(BasicLevel.ERROR, err);
                throw new WSServiceException(err);
            }
        }

        ref.add(new StringRefAddr(REF_WSDD, serializeWSDD(base)));

        // Add a ref for wsdl-file location
        // first add URL of alternate WSDL if possible
        if (sr.getAlternateWsdlURL() != null) {
            logger.log(BasicLevel.DEBUG, "add ref on '" + sr.getAlternateWsdlURL() + "'");
            ref.add(new StringRefAddr(REF_WSDL_URL, sr.getAlternateWsdlURL().toString()));
        } else if (sr.getWSDLFile() != null) {
            String wsdl = sr.getWSDLFile().getName();

            if (wsdl.indexOf(WEB_WSDL_DIR) != -1) {
                wsdl = wsdl.substring(WEB_WSDL_DIR.length());

            } else if (wsdl.indexOf(EJB_WSDL_DIR) != -1) {
                wsdl = wsdl.substring(EJB_WSDL_DIR.length());

            }
            logger.log(BasicLevel.DEBUG, "add ref on '" + wsdl + "'");
            ref.add(new StringRefAddr(REF_WSDL, wsdl));
        }

        // Add ref for service-qname value
        if (sr.getServiceQName() != null) {
            QName  qn = sr.getServiceQName();
            String ns = qn.getNamespaceURI();
            String lp = qn.getLocalPart();
            ref.add(new StringRefAddr(REF_SERVICE_QNAME_NS, ns));
            ref.add(new StringRefAddr(REF_SERVICE_QNAME_LP, lp));
        }

        // manage port-component-link
        String list = "";
        for (Iterator i = sr.getPortComponentRefs().iterator(); i.hasNext();) {
            PortComponentRefDesc pcr = (PortComponentRefDesc) i.next();
            PortComponentDesc pref = pcr.getPortComponentDesc();
            logger.log(BasicLevel.DEBUG, "port-component-link ? for " + pcr.getSEI());
            // port has a link on a local port
            if (pref != null) {
                // lookup endpoint URL value
                URL endpoint = null;
                try {
                    Context ic = new InitialContext();
                    endpoint = (URL) ic.lookup(pref.getName());
                } catch (NamingException ne) {
                    throw new WSServiceException("Cannot find updated endpoint for port-component '" + pref.getName() + "'", ne);
                }
                // wsdl port's name
                ref.add(new StringRefAddr(REF_URL_PREFIX + pref.getQName().getLocalPart(),
                                          endpoint.toString()));
                logger.log(BasicLevel.DEBUG, "port <" + pref.getQName().getLocalPart() + "> endpoint " + endpoint);
                if (!list.equals("")) {
                    list += LIST_SEP;
                }
                list += pref.getQName().getLocalPart();
            }
        }
        // store only when we have links
        if (!list.equals("")) {
            ref.add(new StringRefAddr(REF_PORT_LINK_LIST, list));
        }

        return ref;
    }

    /**
     * Create the Axis Service Implementation class name
     * from the QName (if generated interface found).
     *
     * @param sr the service-ref storing service informations
     *
     * @return the fully qualified Service implementation classname.
     */
    private String createServiceClassname(ServiceRefDesc sr) {

        String intfName = sr.getServiceInterface().getName();

        if (intfName.equals("javax.xml.rpc.Service")) {
            // no generated class
            return JONAS_SERVICE_CLASSNAME;
        } else {
            // generated class
            QName qn = sr.getServiceQName();
            MappingFile mf = sr.getMappingFile();

            String p = (String) mf.getMappings().get(qn.getNamespaceURI());

            String classname = "";
            if (p != null) {
                classname = p + "." + Utils.xmlNameToJavaClass(qn.getLocalPart()) + "Locator";
            } else {
                classname = Utils.xmlNameToJavaClass(qn.getLocalPart()) + "Locator";
            }

            return classname;
        }

    }

    /**
     * Write the WSDDDeployment desc of Axis in a String.
     * @param d the wsdd file
     * @return a String containing wsdd content.
     * @throws WSServiceException if IOException occurs during serialization
     */
    private String serializeWSDD(WSDDDeployment d) throws WSServiceException {

        StringWriter sw = new StringWriter();
        SerializationContext ctx = new SerializationContext(sw);

        try {
            d.writeToContext(ctx);
        } catch (IOException ioe) {
            String err = i18n.getMessage("JAxisServiceFactory.serializeWSDD.serFail");
            logger.log(BasicLevel.ERROR, err);
            throw new WSServiceException(err);

        }

        return sw.toString();
    }

    /**
     * Load the Axis Deployment Descriptor from the ClassLoader.
     *
     * @param filename the resource name to load
     * @param cl The ClassLoader used to load the WSDD
     *
     * @return the Axis Deployment Descriptor from the ClassLoader.
     *
     * @throws WSServiceException if resource could not be found or if XML Document cannot be created
     */
    private WSDDDeployment loadAxisDeployment(String filename, ClassLoader cl)
        throws WSServiceException {

        InputStream is = cl.getResourceAsStream(filename);

        //logger.log(BasicLevel.DEBUG, "classloader : " + cl);

        if (is == null) {
            // exception !
            String err = i18n.getMessage("JAxisServiceFactory.loadAxisDeployment.configNotFound",
                                         filename);
            logger.log(BasicLevel.ERROR, err);
            throw new WSServiceException(err);
        }

        WSDDDeployment deployment = null;
        try {
            Document deployDoc = XMLUtils.newDocument(is);
            deployment = new WSDDDeployment(deployDoc.getDocumentElement());
        } catch (Exception e) {
            String err = i18n.getMessage("JAxisServiceFactory.loadAxisDeployment.docCreation",
                                         filename);
            logger.log(BasicLevel.ERROR, err);

            throw new WSServiceException(err, e);
        }

        return deployment;
    }

    /**
     * Create a new instance of Axis Service (or sub classes) from a Reference object.
     *
     * @param refObject the Reference to create an instance from
     * @param name Object Name
     * @param nameCtx Context
     * @param env the environnment
     *
     * @return A Service instance configured
     *
     * @throws Exception when instanciation fail (check if this is a correct behavior !!!)
     */
    public Object getObjectInstance(Object refObject,
                                    Name name,
                                    Context nameCtx,
                                    Hashtable env) throws Exception {

        JService instance = null;
        Object proxy = null;

        if (refObject instanceof Reference) {
            Reference ref = (Reference) refObject;

            // get ClassLoader
            ClassLoader cl = Thread.currentThread().getContextClassLoader();

            // get Class
            Class serviceClass = cl.loadClass(ref.getClassName());

            // new Service
            StringRefAddr qnameNS = (StringRefAddr) ref.get(REF_SERVICE_QNAME_NS);
            StringRefAddr qnameLP = (StringRefAddr) ref.get(REF_SERVICE_QNAME_LP);

            Map updatedURLs = getUpdatedURLs(ref);

            if (JONAS_SERVICE_CLASSNAME.equals(serviceClass.getName())) {
                // default service class

                logger.log(BasicLevel.DEBUG, "default service class");

                QName qname = null;
                if ((qnameNS != null) && (qnameLP != null)) {
                    String ns = (String) qnameNS.getContent();
                    String lp = (String) qnameLP.getContent();
                    qname = new QName(ns, lp);
                }

                StringRefAddr wsdlRef = (StringRefAddr) ref.get(REF_WSDL);
                StringRefAddr wsdlURLRef = (StringRefAddr) ref.get(REF_WSDL_URL);
                InputStream is = null;
                if (wsdlURLRef != null) {
                    String wsdl = (String) wsdlURLRef.getContent();
                    logger.log(BasicLevel.DEBUG, "Load WSDL URL Ref : " + wsdl);
                    URL url = new URL(wsdl);
                    is = url.openStream();
                }
                // if WSDL not loaded
                if (wsdlRef != null && is == null) {
                    // wsdl file specified
                    String wsdl = (String) wsdlRef.getContent();
                    logger.log(BasicLevel.DEBUG, "Load WSDL Ref : " + wsdl);
                    is = cl.getResourceAsStream(wsdl);

                    if (is == null) {
                        // try to load from WEB-INF/
                        logger.log(BasicLevel.DEBUG, "Within Client Container try WEB-INF/" + wsdl);
                        is = cl.getResourceAsStream("WEB-INF/wsdl/" + wsdl);
                    }
                    if (is == null) {
                        // try to load from META-INF/
                        logger.log(BasicLevel.DEBUG, "Within Client Container try META-INF/" + wsdl);
                        is = cl.getResourceAsStream("META-INF/wsdl/" + wsdl);
                    }
                }

                if ((qname == null) && (is == null)) {
                    logger.log(BasicLevel.DEBUG, "Create a new Service instance without params");
                    // no param instance
                    instance = (JService) serviceClass.newInstance();

                } else if ((qname != null) && (is == null)) {
                    logger.log(BasicLevel.DEBUG, "Create a new Service instance with only a QName " + qname);
                    // qname only
                    Constructor ctr = serviceClass.getConstructor(new Class[] {javax.xml.namespace.QName.class});
                    instance = (JService) ctr.newInstance(new Object[] {qname});

                } else if ((qname != null) && (is != null)) {
                    logger.log(BasicLevel.DEBUG, "Create a new Service instance with QName + WSDL " + qname);
                    // qname + wsdl
                    Constructor ctr = serviceClass.getConstructor(new Class[] {org.apache.axis.wsdl.gen.Parser.class,
                                                                               javax.xml.namespace.QName.class});
                    Parser uptodateParser = updateParser(is, updatedURLs, qname);
                    instance = (JService) ctr.newInstance(new Object[] {uptodateParser, qname});

                } else {
                    logger.log(BasicLevel.DEBUG, "Create a new Service instance with WSDL only.");
                    // wsdl
                    Constructor ctr = serviceClass.getConstructor(new Class[] {org.apache.axis.wsdl.gen.Parser.class,
                                                                               javax.xml.namespace.QName.class});
                    Parser uptodateParser = updateParser(is, updatedURLs, qname);
                    instance = (JService) ctr.newInstance(new Object[] {uptodateParser, qname});
                }

            } else {
                logger.log(BasicLevel.DEBUG, "Create a new Generated Service instance");
                // generated class
                instance = (JService) serviceClass.newInstance();
                if (!updatedURLs.isEmpty()) {
                    java.lang.reflect.Method m = serviceClass.getMethod("setEndpointAddress", new Class[] {String.class, String.class});
                    for (Iterator i = updatedURLs.keySet().iterator(); i.hasNext();) {
                        String port = (String) i.next();
                        String endpoint = (String) updatedURLs.get(port);
                        m.invoke(instance, new Object[] {port, endpoint});
                    }
                }
            }

            EngineConfiguration ec = getConfiguration(ref);

            // configure the instance
            instance.setEngine(new AxisClient(ec));

            JServiceProxy handler = new JServiceProxy(instance);
            Class[] serviceInterfaces = serviceClass.getInterfaces();
            Class[] interfaces = new Class[serviceInterfaces.length + 1];
            for (int i = 0; i < serviceInterfaces.length; i++) {
                interfaces[i] = serviceInterfaces[i];
            }
            interfaces[serviceInterfaces.length] = javax.xml.rpc.Service.class;
            proxy = Proxy.newProxyInstance(cl, interfaces, handler);

        }

        return proxy;
    }

    /**
     * Use the Parser interface of Axis to Update the
     * port's location attribute with new URLs.
     *
     * @param is the WSDL as an InputStream
     * @param urls the map of updated URLs (port name -> URL)
     * @param sQName the service containing port to be updated
     *
     * @return an Axis Parser updated corresponding to the given WSDL InputStream
     *
     * @throws Exception when parsing fails
     */
    private Parser updateParser(InputStream is, Map urls, QName sQName) throws Exception {

        // load WSDL into XML Document
        Document doc = XMLUtils.newDocument(is);

        Parser p = new Parser();
        p.run(null, doc);

        ServiceEntry serviceEntry = p.getSymbolTable().getServiceEntry(sQName);
        if (serviceEntry != null) {
            javax.wsdl.Service service = serviceEntry.getService();
            // Start update ports
            for (Iterator i = urls.keySet().iterator(); i.hasNext();) {
                String name = (String) i.next();

                Port port = service.getPort(name);
                String url = (String) urls.get(name);
                setEndpointURL(port, url);
            }

        }

        return p;
    }

    /**
     * Set the port location value in WSDL Definition object.
     *
     * @param port the port to be updated
     * @param url the new location value
     */
    private void setEndpointURL(Port port, String url) {
        List extensions = port.getExtensibilityElements();
        for (Iterator i = extensions.iterator(); i.hasNext();) {
            ExtensibilityElement ext = (ExtensibilityElement) i.next();
            if (ext instanceof SOAPAddress) {
                SOAPAddress addr = (SOAPAddress) ext;
                addr.setLocationURI(url);
            }
        }
    }

    /**
     * Return a map of PortNames -> URL. Used when port component link are found.
     *
     * @param ref the Reference where information is stored
     *
     * @return a map of PortNames -> URL. Used when port component link are found.
     */
    private Map getUpdatedURLs(Reference ref) {
        Map urls = new Hashtable();

        // get the list of port to update
        StringRefAddr listAddr = (StringRefAddr) ref.get(REF_PORT_LINK_LIST);
        if (listAddr != null) {
            String[] names = ((String) listAddr.getContent()).split(LIST_SEP);
            for (int i = 0; i < names.length; i++) {
                // get the updated url
                StringRefAddr urlAddr = (StringRefAddr) ref.get(REF_URL_PREFIX + names[i]);
                if (urlAddr != null) {
                    urls.put(names[i], urlAddr.getContent());
                }
            }
        }
        return urls;
    }

    /**
     * Return the EngineConfiguration object found in the Reference
     * @param ref The reference to an Axis Service instance
     * @return the EngineConfiguration object found in the Reference
     * @throws Exception when content of reference is not a valid WSDD
     */
    private EngineConfiguration getConfiguration(Reference ref) throws Exception {
        // create a Configuration Object
        String conf = (String) ref.get(REF_WSDD).getContent();
        logger.log(BasicLevel.DEBUG, "loaded configuration : " + conf);
        return new XMLStringProvider(conf);
    }


}
