/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999-2004 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: QSUpdateServiceWSDLHandler.java,v 1.1 2004/09/10 11:46:49 sauthieg Exp $
 * --------------------------------------------------------------------------
 */
package org.objectweb.jonas.ws.axis;

import java.io.PrintWriter;
import java.net.HttpURLConnection;
import java.util.Iterator;
import java.util.List;

import javax.servlet.http.HttpServletResponse;
import javax.wsdl.Definition;
import javax.wsdl.Port;
import javax.wsdl.Service;
import javax.wsdl.WSDLException;
import javax.wsdl.extensions.ExtensibilityElement;
import javax.wsdl.extensions.soap.SOAPAddress;
import javax.wsdl.factory.WSDLFactory;
import javax.xml.namespace.QName;

import org.w3c.dom.Document;

import org.apache.axis.AxisFault;
import org.apache.axis.Constants;
import org.apache.axis.MessageContext;
import org.apache.axis.WSDDEngineConfiguration;
import org.apache.axis.deployment.wsdd.WSDDService;
import org.apache.axis.i18n.Messages;
import org.apache.axis.server.AxisServer;
import org.apache.axis.transport.http.AbstractQueryStringHandler;
import org.apache.axis.transport.http.HTTPConstants;
import org.apache.axis.utils.XMLUtils;

/**
 * Update the wsdlFile specified in wsdd with the good URLs for endpoints.
 * responds to the ?ServiceWSDL
 * @author Guillaume Sauthier
 */
public class QSUpdateServiceWSDLHandler extends AbstractQueryStringHandler {

    /**
     * SOAP NS URI
     */
    private static final String NS_URI_SOAP = "http://schemas.xmlsoap.org/wsdl/soap/";

    /**
     * soap:address Type QName
     */
    private static final QName QNAME_SOAP_ADDRESS = new QName(NS_URI_SOAP, "address");

    /**
     * Performs the action associated with this particular query string handler.
     * @param msgContext a MessageContext object containing message context
     *        information for this query string handler.
     * @throws AxisFault if an error occurs.
     */
    public void invoke(MessageContext msgContext) throws AxisFault {
        // Obtain objects relevant to the task at hand from the provided
        // MessageContext's bag.
        configureFromContext(msgContext);

        AxisServer engine = (AxisServer) msgContext.getProperty(HTTPConstants.PLUGIN_ENGINE);
        PrintWriter writer = (PrintWriter) msgContext.getProperty(HTTPConstants.PLUGIN_WRITER);
        HttpServletResponse response = (HttpServletResponse) msgContext
                .getProperty(HTTPConstants.MC_HTTP_SERVLETRESPONSE);

        try {
            // get the WSDL file
            engine.generateWSDL(msgContext);

            // as a Document
            Document doc = (Document) msgContext.getProperty("WSDL");

            if (doc != null) {
                // update WSDL
                Document up2date = updatePortLocation(engine, doc);
                response.setContentType("text/xml; charset=" + XMLUtils.getEncoding().toLowerCase());
                reportWSDL(up2date, writer);
            } else {
                // report Error
                if (log.isDebugEnabled()) {
                    log.debug("processWsdlRequest: failed to create WSDL");
                }
                reportNoWSDL(response, writer, "noWSDL02", null);
            }
        } catch (AxisFault axisFault) {
            //the no-service fault is mapped to a no-wsdl error
            if (axisFault.getFaultCode().equals(Constants.QNAME_NO_SERVICE_FAULT_CODE)) {
                //which we log
                processAxisFault(axisFault);

                //then report under a 404 error
                response.setStatus(HttpURLConnection.HTTP_NOT_FOUND);
                reportNoWSDL(response, writer, "noWSDL01", axisFault);
            } else {
                //all other faults get thrown
                throw axisFault;
            }
        }
    }

    /**
     * @param engine Engine storing Updated Configuration
     * @param doc Document to update
     * @throws AxisFault When Document cannot be parsed as a WSDL Definition instance
     */
    private Document updatePortLocation(AxisServer engine, Document doc) throws AxisFault {
        log.debug("Entering updatePortLocation");
        try {
            // get Definition from Document
            Definition def = WSDLFactory.newInstance().newWSDLReader().readWSDL(null, doc);
            // get Engine Configuration
            WSDDEngineConfiguration conf = (WSDDEngineConfiguration) engine.getConfig();
            WSDDService[] services = conf.getDeployment().getServices();
            // for each service/endpoint
            for (int i = 0; i < services.length; i++) {
                // get Updated endpoint URL
                String url = services[i].getServiceDesc().getEndpointURL();
                QName qname = services[i].getQName();
                Service s = (Service) def.getServices().values().iterator().next();
                Port p = s.getPort(qname.getLocalPart());
                
                // Set the updated soap:address address
                List ext = p.getExtensibilityElements();
                for (Iterator it = ext.iterator(); it.hasNext();) {
                    ExtensibilityElement element = (ExtensibilityElement) it.next();
                    if (element.getElementType().equals(QNAME_SOAP_ADDRESS)) {
                        SOAPAddress sa = (SOAPAddress) element;
                        sa.setLocationURI(url);
                    }
                }
            }
            return WSDLFactory.newInstance().newWSDLWriter().getDocument(def);
        } catch (WSDLException wsdle) {
            throw new AxisFault("Cannot read WSDL Document", wsdle);
        }
    }

    /**
     * Reports WSDL
     * @param doc Document to write
     * @param writer Servlet Writer to use
     */
    public void reportWSDL(Document doc, PrintWriter writer) {
        XMLUtils.PrettyDocumentToWriter(doc, writer);
    }

    /**
     * Reports that we have no WSDL
     * @param res HttpServletResponse
     * @param writer PrintWriter
     * @param moreDetailCode optional name of a message to provide more detail
     * @param axisFault optional fault string, for extra info at debug time only
     */
    public void reportNoWSDL(HttpServletResponse res, PrintWriter writer, String moreDetailCode, AxisFault axisFault) {
        res.setStatus(HttpURLConnection.HTTP_NOT_FOUND);
        res.setContentType("text/html");

        writer.println("<h2>" + Messages.getMessage("error00") + "</h2>");
        writer.println("<p>" + Messages.getMessage("noWSDL00") + "</p>");

        if (moreDetailCode != null) {
            writer.println("<p>" + Messages.getMessage(moreDetailCode) + "</p>");
        }

        if (axisFault != null && isDevelopment()) {
            //dev systems only give fault dumps
            writeFault(writer, axisFault);
        }
    }

}