/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: JMdbEndpointFactory.java,v 1.12 2004/09/17 22:33:32 ehardesty Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.ArrayList;
import java.util.List;
import java.util.LinkedList;
import java.util.ListIterator;

import javax.ejb.EJBException;
import javax.ejb.MessageDrivenBean;
import javax.ejb.MessageDrivenContext;
import javax.ejb.Timer;
import javax.ejb.TimerService;
import javax.naming.Context;
import javax.resource.spi.ActivationSpec;
import javax.resource.spi.UnavailableException;
import javax.resource.spi.endpoint.MessageEndpoint;
import javax.resource.spi.endpoint.MessageEndpointFactory;
import javax.transaction.SystemException;
import javax.transaction.xa.XAResource;

import org.objectweb.jonas_ejb.deployment.api.MessageDrivenDesc;
import org.objectweb.jonas_ejb.deployment.api.MethodDesc;
import org.objectweb.jonas_ejb.deployment.api.ActivationConfigPropertyDesc;

import org.objectweb.jonas.resource.Rar;
import org.objectweb.jonas.resource.ResourceServiceImpl;

import org.objectweb.util.monolog.api.BasicLevel;

/**
 * This class is a factory for a Message Driven Bean Endpoints There is one such
 * class per MDB class. Contains all information related to the bean
 * @author Eric Hardesty
 */
public class JMdbEndpointFactory extends JFactory implements MessageEndpointFactory {

    /**
     * pool of JMessageEndpoint objects
     */
    private List endpool = new ArrayList();

    protected int instanceCount = 0;

    // initial value for pool size
    protected int minPoolSize = 0;

    // nb max of instances in pool
    protected int maxCacheSize = 0;

    private ActivationSpec as = null;

    private String msglistenType = null;

    private Rar rar = null;

    private boolean activated = false;

    private static String JMESSAGEENDPOINT_CLASS = "org.objectweb.jonas_ejb.container.JMessageEndpointIntf";

    private static String MESSAGEENDPOINT_CLASS = "javax.resource.spi.endpoint.MessageEndpoint";

    private static String MESSAGEDRIVENCONTEXT_CLASS = "javax.ejb.MessageDrivenContext";

    /**
     * Constructor
     * @param dd Message Driven Descriptor
     * @param cont Container where this bean is defined
     * @param _as ActivationSpec to link the Container to
     */
    public JMdbEndpointFactory(MessageDrivenDesc dd, JContainer cont, ActivationSpec _as) {
        super(dd, cont);
        String dest = dd.getDestinationJndiName();
        // Set the AS properties and validate the required config properties
        List acpl = null;
        if (dd.getMdActivationConfigDesc() != null) {
            acpl = dd.getMdActivationConfigDesc().getActivationConfigPropertyList();
        }
        List jAcpl = null;
        if (dd.getJonasMdActivationConfigDesc() != null) {
            jAcpl = dd.getJonasMdActivationConfigDesc().getActivationConfigPropertyList();
        }
        mdbEFInit(dd, dest, acpl, jAcpl, _as);
    }

    /**
     * Constructor
     * @param dd Message Driven Descriptor
     * @param cont Container where this bean is defined
     * @param _as ActivationSpec to link the Container to
     */
    public JMdbEndpointFactory(MessageDrivenDesc dd, String destination, JContainer cont, 
                               ActivationSpec _as) {
        super(dd, cont);
        
        // Build Activation spec linked list from variables
        
        List jAcpl = new LinkedList();
        List acpl = null;
        if (dd.getMdActivationConfigDesc() != null) {
            acpl = dd.getMdActivationConfigDesc().getActivationConfigPropertyList();
        }
        buildACL(dd, jAcpl);
        
        mdbEFInit(dd, destination, acpl, jAcpl, _as);
    }
    
    /**
     * Create the ActivationConfigProperty List 
     * @param jacl List to update
     * @param destination String destination to to swap with default value
     */
    private void buildACL(MessageDrivenDesc dd, List jacl) {

        String [] aclNames = {"destination", "destinationType","messageSelector", 
                              "acknowledgeMode", "subscriptionDurability" }; 
        
        ActivationConfigPropertyDesc acp = null;
        
        acp = new ActivationConfigPropertyDesc();
        acp.setActivationConfigPropertyName("destination");
        acp.setActivationConfigPropertyValue(dd.getDestinationJndiName());
        jacl.add(acp);

        acp = new ActivationConfigPropertyDesc();
        acp.setActivationConfigPropertyName("destinationType");
        if (dd.isTopicDestination()) {
            acp.setActivationConfigPropertyValue("javax.jms.Topic");
        } else {
            acp.setActivationConfigPropertyValue("javax.jms.Queue");
        }
        jacl.add(acp);

        acp = new ActivationConfigPropertyDesc();
        acp.setActivationConfigPropertyName("messageSelector");
        acp.setActivationConfigPropertyValue(dd.getSelector());
        jacl.add(acp);

        acp = new ActivationConfigPropertyDesc();
        acp.setActivationConfigPropertyName("acknowledgeMode");
        if (dd.getAcknowledgeMode() == MessageDrivenDesc.AUTO_ACKNOWLEDGE) {
            acp.setActivationConfigPropertyValue("Auto-acknowledge");
        } else {
            acp.setActivationConfigPropertyValue("Dups-ok-acknowledge");
        }
        jacl.add(acp);

        acp = new ActivationConfigPropertyDesc();
        acp.setActivationConfigPropertyName("subscriptionDurability");
        if (dd.getSubscriptionDurability() == MessageDrivenDesc.SUBS_DURABLE) {
            acp.setActivationConfigPropertyValue("Durable");
        } else {
            acp.setActivationConfigPropertyValue("NonDurable");
        }
        jacl.add(acp);

        if (dd.getSubscriptionDurability() == MessageDrivenDesc.SUBS_DURABLE) {
            acp = new ActivationConfigPropertyDesc();
            acp.setActivationConfigPropertyName("subscriptionName");
            acp.setActivationConfigPropertyValue(dd.getEjbName());
            jacl.add(acp);
        }

        // Check existing list to determine if additional items are configured, then pass them along also
        List acpl = null;
        if (dd.getJonasMdActivationConfigDesc() != null) {
            acpl = dd.getJonasMdActivationConfigDesc().getActivationConfigPropertyList();
            
            String acpName = null;
            boolean found = false;
            for (int i = 0; i < acpl.size(); i++) {
                acp = (ActivationConfigPropertyDesc) acpl.get(i);
                acpName = acp.getActivationConfigPropertyName();
                found = false;
                for (int j = 0; j < aclNames.length; j++) {
                    if (acpName.equals(aclNames[j])) {
                        found = true;
                        break;
                    }
                }
                if (!found) {
                    jacl.add(acp);
                }
            }

        }
        
    }
    
    /**
     * Init this endpointfactory
     * @param dd Message Driven Descriptor
     * @param cont Container where this bean is defined
     * @param _as ActivationSpec to link the Container to
     */
    private void mdbEFInit(MessageDrivenDesc dd, String dest, List acpl,
                           List jAcpl, ActivationSpec _as) {

        String ejbName = dd.getEjbName();

        // Check if tx managed by the bean or the container
        txbeanmanaged = dd.isBeanManagedTransaction();

        // Set ActivationSpec
        as = _as;

        minPoolSize = dd.getPoolMin();
        maxCacheSize = dd.getCacheMax();
        if (TraceEjb.isDebugSwapper()) {
            TraceEjb.swapper.log(BasicLevel.DEBUG, " maxCacheSize = " + maxCacheSize + " minPoolSize = " + minPoolSize);
        }

        // Get the associated Resource Adapter and messagelistenerType
        rar = ResourceServiceImpl.getRar(dest);
        if (rar == null) {
            TraceEjb.mdb.log(BasicLevel.ERROR, "cannot retrieve associated Resource Adapter ");
            throw new EJBException("cannot retrieve associated Resource Adapter ");
        }
        msglistenType = rar.getInterface(dest);

        try {
            rar.configureAS(as, acpl, jAcpl, dest, ejbName);
        } catch (Exception ex) {
            TraceEjb.mdb.log(BasicLevel.ERROR, "cannot configure activationspec " + ex);
            ex.printStackTrace();
            throw new EJBException("cannot configure activationspec ", ex);
        }

        // pre-allocate a set of Endpoints
        synchronized (endpool) {
            for (int i = 0; i < minPoolSize; i++) {
                JMessageEndpoint ep = null;
                try {
                    ep = createNewInstance();
                    endpool.add(ep);
                } catch (Exception e) {
                    TraceEjb.mdb.log(BasicLevel.ERROR, "cannot init pool of instances ");
                    throw new EJBException("cannot init pool of instances ", e);
                }
            }
        }
        if (minPoolSize != 0) {
            TraceEjb.mdb.log(BasicLevel.INFO, "pre-allocate a set of " + minPoolSize
                    + " message driven bean  instances");
        }

        activated = true;
        // EndpointActivation
        try {
            rar.getResourceAdapter().endpointActivation(this, as);
        } catch (Exception ex) {
            activated = false;
            TraceEjb.mdb.log(BasicLevel.ERROR, "cannot activate endpoint " + ex);
            ex.printStackTrace();
            throw new EJBException("cannot activate endpoint ", ex);
        }
    }

    // ---------------------------------------------------------------
    // Specific BeanFactory implementation
    // ---------------------------------------------------------------

    /**
     * @return the size of the EndpointPool
     */
    public int getPoolSize() {
        return endpool.size();
    }

    /**
     * stop this EJB. call deactivate on the Endpoint Stop the threads and
     * remove the beans
     */
    public void stop() {
        TraceEjb.mdb.log(BasicLevel.DEBUG, "");
        try {
            // Deactivate the endpoint
            rar.getResourceAdapter().endpointDeactivation(this, as);
            // Loop thru the endpool and release them
            synchronized (endpool) {
                if (TraceEjb.isDebugSwapper()) {
                    TraceEjb.swapper.log(BasicLevel.DEBUG, "stopping " + this);
                }
                JMessageEndpoint ep = null;
                while (endpool.size() > 0) {
                    ep = (JMessageEndpoint) endpool.remove(0);
                    instanceCount--;
                    try {
                        ep.mdb.ejbRemove();
                    } catch (Exception e) {
                        TraceEjb.logger.log(BasicLevel.ERROR, "stop: Cannot remove mdb: "+ep.mdb);
                    }
                }
            }
        } catch (Exception e) {
            TraceEjb.logger.log(BasicLevel.WARN, "stop: Cannot deactivate the endpoint");
        }
    }

    /**
     * synchronize bean instances if needed
     */
    public void sync() {
    }

    /**
     * @return the home if exist
     */
    public JHome getHome() {
        return null;
    }

    /**
     * @return the local home if exist
     */
    public JLocalHome getLocalHome() {
        return null;
    }

    // ---------------------------------------------------------------
    // MessageEndpointFactory implementation
    // ---------------------------------------------------------------

    /**
     * @return the MessageEndpoint to deliver messages to
     */
    public MessageEndpoint createEndpoint(XAResource xaResource) throws UnavailableException {
        TraceEjb.mdb.log(BasicLevel.DEBUG, "");

        if (!activated) {
            TraceEjb.mdb.log(BasicLevel.ERROR, "mdb is not usuable either initial deployment or undeployment ");
            throw new UnavailableException("mdb is not usuable either initial deployment or undeployment ");
        }

        JMessageEndpoint ep = null;
        try {
            ep = getNewInstance(xaResource);
        } catch (Exception ex) {
            TraceEjb.mdb.log(BasicLevel.ERROR, "cannot create an endpoint ");
            throw new UnavailableException("cannot create an endpoint ", ex);
        }
        return (MessageEndpoint) ep.mep;
    }

    /**
     * @return boolean whether the specified method is transacted
     */
    public boolean isDeliveryTransacted(Method method) throws NoSuchMethodException {
        int txAttribute = 0;
        try {
            txAttribute = dd.getMethodDesc(method).getTxAttribute();
        } catch (Exception ex) {
            TraceEjb.mdb.log(BasicLevel.ERROR, "No such method exists. " + method.getName());
            throw new NoSuchMethodException("No such method exists. " + method.getName());
        }
        return txAttribute == MethodDesc.TX_REQUIRED ? true : false;
    }

    // ---------------------------------------------------------------
    // Other methods
    // ---------------------------------------------------------------

    /**
     * Return an MessageEndpoint from the pool. If pool is empty, creates a new
     * one.
     * @return an MessageEndpoint from the pool.
     * @exception Exception - if an application server fails to return an
     *            MessageEndpoint out of its pool.
     */
    public JMessageEndpoint getMessageEndpoint() throws Exception {
        TraceEjb.mdb.log(BasicLevel.DEBUG, "");

        return getNewInstance(null);
    }

    /**
     * put the JMessageEndpoint back to the pool
     * @param ep the MessageEndpoint
     */
    public void releaseEndpoint(JMessageEndpoint ep) {
        TraceEjb.mdb.log(BasicLevel.DEBUG, "EP:" + ep);

        ep.setReleasedState(true);
        synchronized (endpool) {
            endpool.add(ep);
        }
        if (TraceEjb.isDebugJms()) {
            TraceEjb.mdb.log(BasicLevel.DEBUG, "nb instances " + getCacheSize());
            TraceEjb.mdb.log(BasicLevel.DEBUG, "nb free cached instances " + getPoolSize());
        }

    }

    // ---------------------------------------------------------------
    // other public methods
    // ---------------------------------------------------------------

    /**
     * Obtains the TimerService associated for this Bean
     * @return a JTimerService instance.
     */
    public TimerService getTimerService() {
        if (myTimerService == null) {
            // TODO : Check that instance implements TimedObject ?
            myTimerService = new JTimerService(this);
        }
        return myTimerService;
    }

    /**
     * @return min pool size for Jmx
     */
    public int getMinPoolSize() {
        return minPoolSize;
    }

    /**
     * @return max cache size for Jmx
     */
    public int getMaxCacheSize() {
        return maxCacheSize;
    }

    /**
     * @return current cache size ( = nb of instance created) for Jmx
     */
    public int getCacheSize() {
        return instanceCount;
    }

    /**
     * @return the Transaction Attribute
     */
    public int getTransactionAttribute() {
        return ((MessageDrivenDesc) dd).getTxAttribute();
    }

    /**
     * For Message Driven Beans, only 2 cases are possible.
     * @param rctx The Request Context
     */
    public void checkTransaction(RequestCtx rctx) {
        if (rctx.txAttr == MethodDesc.TX_REQUIRED) {
            try {
                if (txbeanmanaged) {
                    if (tm.getTransaction() == null) {
                        TraceEjb.logger.log(BasicLevel.ERROR, "No transaction and need one");
                        return;
                    }
                } else {
                    if (tm.getTransaction() == null) {
                        tm.begin();
                    }
                }
                rctx.mustCommit = true;
                rctx.currTx = tm.getTransaction();
            } catch (Exception e) {
                // No exception raised in case of MDB
                TraceEjb.logger.log(BasicLevel.ERROR, "cannot start tx:" + e);
                return;
            }
        } else {
            try {
                rctx.currTx = tm.getTransaction();
                if (rctx.currTx != null) {
                    rctx.clientTx = tm.suspend();
                    rctx.currTx = null;
                }
            } catch (SystemException e) {
                TraceEjb.logger.log(BasicLevel.ERROR, "cannot suspend transaction:\n", e);
                return;
            }
        }
    }

    /**
     * Reduce number of instances in memory in the free list we reduce to the
     * minPoolSize
     */
    public void reduceCache() {
        TraceEjb.swapper.log(BasicLevel.DEBUG, "");
        // reduce the pool to the minPoolSize
        int poolsz = minPoolSize;
        synchronized (endpool) {
            if (TraceEjb.isDebugSwapper()) {
                TraceEjb.swapper.log(BasicLevel.DEBUG, "try to reduce " + endpool.size() + " to " + poolsz);
            }
            while (endpool.size() > poolsz) {
                ListIterator i = endpool.listIterator();
                if (i.hasNext()) {
                    i.next();
                    i.remove();
                    instanceCount--;
                }
            }
        }
        if (TraceEjb.isDebugSwapper()) {
            TraceEjb.swapper.log(BasicLevel.DEBUG, "cacheSize= " + getCacheSize());
        }

    }

    /**
     * Notify a timeout for this bean
     * @param timer timer whose expiration caused this notification.
     */
    public void notifyTimeout(Timer timer) {
        TraceEjb.mdb.log(BasicLevel.DEBUG, "");

        // We need an instance from the pool to process the timeout.
        JMessageEndpoint ep = null;
        try {
            ep = getNewInstance(null);
        } catch (Exception e) {
            TraceEjb.logger.log(BasicLevel.ERROR, "exception:" + e);
            throw new EJBException("Cannot deliver the timeout: " + e);
        }

        // deliver the timeout to the bean
        ep.deliverTimeout(timer);

        // release the instance
        releaseEndpoint(ep);
    }

    // ---------------------------------------------------------------
    // private methods
    // ---------------------------------------------------------------

    /**
     * return a new instance of the bean. Try to get one from the pool, and
     * create a new one if the pool is empty.
     */
    private JMessageEndpoint getNewInstance(XAResource xaResource) throws Exception {
        TraceEjb.mdb.log(BasicLevel.DEBUG, "Factory: " + this + " XAResource: " + xaResource);

        // try to get one from the Pool
        JMessageEndpoint ep = null;

        // try to find a free entry in the pool
        synchronized (endpool) {
            if (!endpool.isEmpty()) {
                try {
                    ep = (JMessageEndpoint) endpool.remove(0);
                } catch (Exception ex) {
                    // This should never happen
                    TraceEjb.logger.log(BasicLevel.ERROR, "Exception:" + ex);
                    throw new EJBException("Cannot get an instance from the pool");
                }
            } else {
                TraceEjb.mdb.log(BasicLevel.DEBUG, "pool is empty");
                if (maxCacheSize == 0 || instanceCount < maxCacheSize) {
                    // Pool has free slots, create a new MessageEndpoint object
                    try {
                        ep = createNewInstance();
                    } catch (Exception e) {
                        TraceEjb.logger.log(BasicLevel.ERROR, "exception:" + e);
                        throw new EJBException("Cannot create a new instance");
                    }
                } else {
                    EJBException ex = new EJBException("Cannot create a new instance, no room in pool");
                    TraceEjb.logger.log(BasicLevel.ERROR, "exception:" + ex);
                    throw ex;
                }

            }
            if (TraceEjb.isDebugSwapper()) {
                TraceEjb.swapper.log(BasicLevel.DEBUG, "nb instances " + getCacheSize());
            }
            ep.setXAResource(xaResource);
            ep.setReleasedState(false);
            TraceEjb.mdb.log(BasicLevel.DEBUG, " Returning " + ep);
            return ep;
        }
    }

    /**
     * Create a new instance of the bean
     */
    private JMessageEndpoint createNewInstance() throws Exception {
        TraceEjb.mdb.log(BasicLevel.DEBUG, "");
        JMessageEndpointProxy epProxy = null;
        MessageEndpoint ep = null;
        JMessageEndpoint jep = null;
        MessageDrivenDesc mdd = (MessageDrivenDesc) dd;
        ClassLoader cls = myClassLoader();

        // Set ContextClassLoader with the ejbclassloader.
        // This is necessary in case ejbCreate calls another bean in the same
        // jar.
        Thread.currentThread().setContextClassLoader(cls);

        // Creates the new instance
        MessageDrivenBean mdb = null;
        try {
            mdb = (MessageDrivenBean) beanclass.newInstance();
        } catch (Exception e) {
            TraceEjb.logger.log(BasicLevel.ERROR, "failed to create instance:", e);
            throw new EJBException("Container failed to create instance of Message Driven Bean", e);
        }

        // Instanciates a new JMessageEndpoint object
        jep = new JMessageEndpoint(this, mdb);
        epProxy = new JMessageEndpointProxy(this, mdb, jep);
        Class endpointClass = cls.loadClass(MESSAGEENDPOINT_CLASS);
        Class msgListenerClass = cls.loadClass(msglistenType);
        ep = (MessageEndpoint) Proxy.newProxyInstance(cls, new Class[] {endpointClass, msgListenerClass}, epProxy);

        jep.setProxy(ep);
        // starts the bean instance: setMessageDrivenContext() + ejbCreate()
        // see EJB spec. 2.0 page 322.
        // Both operations must be called with the correct ComponentContext
        Context ctxsave = setComponentContext();
        mdb.setMessageDrivenContext((MessageDrivenContext) jep);
        try {
            beanclass.getMethod("ejbCreate", null).invoke(mdb, null);
        } catch (Exception e) {
            TraceEjb.logger.log(BasicLevel.ERROR, "cannot call ejbCreate on message driven bean instance ", e);
            throw new EJBException(" Container fails to call ejbCreate on message driven bean instance", e);
        } finally {
            resetComponentContext(ctxsave);
        }

        synchronized (endpool) {
            instanceCount++;
        }
        return jep;
    }

}
