/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999-2004 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: BeanSources.java,v 1.29 2004/09/21 15:05:46 joaninh Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.genic;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.Properties;
import java.util.StringTokenizer;

import org.apache.velocity.app.VelocityEngine;

import org.objectweb.jonas_ejb.deployment.api.BeanDesc;
import org.objectweb.jonas_ejb.deployment.api.EntityCmp2Desc;
import org.objectweb.jonas_ejb.deployment.api.EntityDesc;
import org.objectweb.jonas_ejb.deployment.api.EntityJdbcCmp1Desc;
import org.objectweb.jonas_ejb.deployment.api.SessionStatelessDesc;
import org.objectweb.jonas_ejb.lib.BeanNaming;
import org.objectweb.jonas_ejb.lib.RdbMappingBuilder;

import org.objectweb.jonas.common.Log;

import org.objectweb.jorm.compiler.api.JormCompilerConfigurator;
import org.objectweb.jorm.compiler.api.JormCompilerParameter;
import org.objectweb.jorm.compiler.lib.JormCompiler;
import org.objectweb.jorm.metainfo.api.Manager;

import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.wrapper.printwriter.LoggerImpl;
import org.objectweb.util.monolog.wrapper.velocity.VelocityLogger;

/**
 * This class allows to generate the sources of: the class that implements the
 * Enterprise bean's remote interface, the class that implements the Enterprise
 * bean's home interface, the class that implements the Enterprise bean's local
 * interface, the class that implements the Enterprise bean's localhome
 * interface, the class of the Entity Handle in case of entity, and the extended
 * class of the Bean for persistence in case of entity with CMP, of a given
 * Enterprise Java Bean.
 * @author Helene Joanin : Initial developer
 * @author Philippe Durieux
 * @author Sebastien Chassande
 */
public class BeanSources {

    /**
     * install.root property name
     */
    private static final String INSTALL_ROOT_PROPERTY = "install.root";

    /**
     * java file suffix
     */
    private static final String JAVA_SUFFIX = ".java";

    /**
     * mapper name prefix
     */
    private static final String RDB_PREFIX = "rdb";

    /**
     * Cluster suffix
     */
    private static final String CLUSTER_SUFFIX = "_Cluster.xml";

    /**
     * EJBObject (Remote). May be null if no Remote interface.
     */
    private String wrpRemoteFileName = null;

    /**
     * EJBLocalObject (Local). May be null if no Local interface.
     */
    private String wrpLocalFileName = null;

    /**
     * Home. May be null if no Home interface.
     */
    private String wrpHomeFileName = null;

    /**
     * Home cluster configuration file. May be null if no Home interface.
     */
    private String wrpHomeClusterConfFileName = null;

    /**
     * LocalHome. May be null if no LocalHome interface.
     */
    private String wrpLocalHomeFileName = null;

    /**
     * Handle (for Entity only)
     */
    private String wrpHandleFileName = null;

    /**
     * Bean (for CMP Entity only)
     */
    private String wrpBeanFileName = null;

    /**
     * Service Endpoint filename
     */
    private String wrpServiceEndpointFileName = null;

    /**
     * Interface for Cmp2 Entity coherence
     */
    private String itfCohCmp2Entity = null;

    /**
     * list of mappers names
     */
    private ArrayList mapperList;

    /**
     * Source objects used to generate classes
     */
    private ArrayList sources;

    /**
     * Java sources that are not Remote
     */
    private ArrayList noRemoteJava;

    /**
     * directory where to place the generated files
     */
    private String outdir = null;

    /**
     * bean descriptor
     */
    private BeanDesc dd = null;

    /**
     * The JORM meta-info manager
     */
    private Manager manager = null;

    /**
     * logger
     */
    private Logger logger = null;

    /**
     * GenIC Constructor
     *
     * @param beanDesc deployment descriptor of the bean
     * @param dirOutputName path of the directory where to place the generated
     *        files (empty string "" if the output directory is the current
     *        directory)
     * @param manager JORM meta_info manager
     * @exception GenICException In error case
     */
    public BeanSources(BeanDesc beanDesc, String dirOutputName, Manager manager) throws GenICException {
        this(beanDesc, dirOutputName);
        this.manager = manager;
    }

    /**
     * GenIC Constructor
     *
     * @param beanDesc deployment descriptor of the bean
     * @param dirOutputName path of the directory where to place the generated
     *        files (empty string "" if the output directory is the current
     *        directory)
     * @exception GenICException In error case
     */
    public BeanSources(BeanDesc beanDesc, String dirOutputName) throws GenICException {

        logger = Log.getLogger(Log.JONAS_GENIC_PREFIX);
        dd = beanDesc;
        outdir = dirOutputName;
        sources = new ArrayList();
        noRemoteJava = new ArrayList();

        // Create and initialize the Velocity Engine
        String jonasRoot = System.getProperty(INSTALL_ROOT_PROPERTY);
        if (jonasRoot == null) {
            throw new GenICException("System property '" + INSTALL_ROOT_PROPERTY + "' not set");
        }
        String path2Tmpl = new String(jonasRoot + File.separatorChar + "templates" + File.separatorChar + "genic");

        VelocityEngine ve = new VelocityEngine();
        ve.setProperty(VelocityEngine.VM_LIBRARY, "");
        ve.setProperty(VelocityEngine.RESOURCE_LOADER, "file");
        ve.setProperty(VelocityEngine.FILE_RESOURCE_LOADER_PATH, path2Tmpl);
        // Velocity logs
        Logger vLogger = Log.getLogger(Log.JONAS_GENIC_VELOCITY_PREFIX);
        if (vLogger.getCurrentIntLevel() != BasicLevel.DEBUG) {
            // No more velocity logs to avoid the creation of an empty file
            // "velocity.log"
            ve.setProperty(VelocityEngine.RUNTIME_LOG_LOGSYSTEM_CLASS, "org.apache.velocity.runtime.log.NullLogSystem");
        } else {
            // Connect the velocity log system to monolog (no more needed)
            ve.setProperty(VelocityEngine.RUNTIME_LOG_LOGSYSTEM, new VelocityLogger(vLogger));
        }
        try {
            ve.init();
        } catch (Exception e) {
            throw new GenICException("Cannot initialize the Velocity engine", e);
        }

        // generated Home
        if (dd.getHomeClass() != null) {
            if (outdir.length() > 0) {
                wrpHomeFileName = outdir + File.separatorChar + BeanNaming.getPath(dd.getFullWrpHomeName());
            } else {
                wrpHomeFileName = dd.getWrpHomeName();
            }
            wrpHomeClusterConfFileName = wrpHomeFileName.concat(CLUSTER_SUFFIX);
            sources.add(new Source(dd, wrpHomeClusterConfFileName, Source.CLUSTER_HOME, ve));
            wrpHomeFileName = wrpHomeFileName.concat(JAVA_SUFFIX);
            sources.add(new Source(dd, wrpHomeFileName, Source.HOME, ve));
            // generated Handle for Remote Entity bean
            if (dd instanceof EntityDesc) {
                if (outdir.length() > 0) {
                    wrpHandleFileName = outdir + File.separatorChar + BeanNaming.getPath(dd.getFullWrpHandleName());
                } else {
                    wrpHandleFileName = dd.getWrpHandleName();
                }
                wrpHandleFileName = wrpHandleFileName.concat(JAVA_SUFFIX);
                noRemoteJava.add(wrpHandleFileName);
                sources.add(new Source(dd, wrpHandleFileName, Source.ENTITY_HANDLE, ve));
            }
        }

        // generated EJBObject (Remote)
        if (dd.getRemoteClass() != null) {
            if (outdir.length() > 0) {
                wrpRemoteFileName = outdir + File.separatorChar + BeanNaming.getPath(dd.getFullWrpRemoteName());
            } else {
                wrpRemoteFileName = dd.getWrpRemoteName();
            }
            wrpRemoteFileName = wrpRemoteFileName.concat(JAVA_SUFFIX);
            sources.add(new Source(dd, wrpRemoteFileName, Source.REMOTE, ve));
        }

        // generated LocalHome
        if (dd.getLocalHomeClass() != null) {
            if (outdir.length() > 0) {
                wrpLocalHomeFileName = outdir + File.separatorChar + BeanNaming.getPath(dd.getFullWrpLocalHomeName());
            } else {
                wrpLocalHomeFileName = dd.getWrpLocalHomeName();
            }
            wrpLocalHomeFileName = wrpLocalHomeFileName.concat(JAVA_SUFFIX);
            noRemoteJava.add(wrpLocalHomeFileName);
            sources.add(new Source(dd, wrpLocalHomeFileName, Source.LOCAL_HOME, ve));
        }

        // generated EJBLocalObject (Local)
        if (dd.getLocalClass() != null) {
            if (outdir.length() > 0) {
                wrpLocalFileName = outdir + File.separatorChar + BeanNaming.getPath(dd.getFullWrpLocalName());
            } else {
                wrpLocalFileName = dd.getWrpLocalName();
            }
            wrpLocalFileName = wrpLocalFileName.concat(JAVA_SUFFIX);
            noRemoteJava.add(wrpLocalFileName);
            sources.add(new Source(dd, wrpLocalFileName, Source.LOCAL, ve));
        }

        // generated ServiceEndpoint
        if (dd instanceof SessionStatelessDesc) {
            SessionStatelessDesc ssd = (SessionStatelessDesc) dd;
            if (ssd.getServiceEndpointClass() != null) {
                if (outdir.length() > 0) {
                    wrpServiceEndpointFileName = outdir + File.separatorChar
                            + BeanNaming.getPath(ssd.getFullWrpServiceEndpointName());
                } else {
                    wrpServiceEndpointFileName = dd.getDerivedBeanName();
                }
                wrpServiceEndpointFileName = wrpServiceEndpointFileName.concat(JAVA_SUFFIX);
                sources.add(new Source(dd, wrpServiceEndpointFileName, Source.SERVICE_ENDPOINT, ve));
            }
        }

        // generated derived bean class for Entity Bean CMP
        if (dd instanceof EntityJdbcCmp1Desc) {
            if (outdir.length() > 0) {
                wrpBeanFileName = outdir + File.separatorChar + BeanNaming.getPath(dd.getFullDerivedBeanName());
            } else {
                wrpBeanFileName = dd.getDerivedBeanName();
            }
            wrpBeanFileName = wrpBeanFileName.concat(JAVA_SUFFIX);
            noRemoteJava.add(wrpBeanFileName);
            sources.add(new Source(dd, wrpBeanFileName, Source.ENTITY_CMP_JDBC, ve));
        }
        if (dd instanceof EntityCmp2Desc) {
            EntityCmp2Desc ecd = (EntityCmp2Desc) dd;
            getAllMapperNames();
            if (outdir.length() > 0) {
                wrpBeanFileName = outdir + File.separatorChar + BeanNaming.getPath(dd.getFullDerivedBeanName())
                        + JAVA_SUFFIX;
                if (ecd.needJormCoherenceHelper()) {
                    itfCohCmp2Entity = outdir + File.separatorChar
                            + BeanNaming.getPath(ecd.getJormCoherenceHelperFQItfName()) + JAVA_SUFFIX;
                }
            } else {
                wrpBeanFileName = BeanNaming.getPath(dd.getDerivedBeanName()) + JAVA_SUFFIX;
                if (ecd.needJormCoherenceHelper()) {
                    itfCohCmp2Entity = BeanNaming.getPath(ecd.getJormCoherenceHelperItfName()) + JAVA_SUFFIX;
                }
            }
            if (ecd.needJormCoherenceHelper()) {
                sources.add(new Source(dd, itfCohCmp2Entity, Source.ITF_COH_CMP2_ENTITY, ve));
                noRemoteJava.add(itfCohCmp2Entity);
            }
            noRemoteJava.add(wrpBeanFileName);
            sources.add(new Source(dd, wrpBeanFileName, Source.ENTITY_CMP_JDBC, ve));
        }
    }

    /**
     * Initializes the list of the mapper names
     */
    private void getAllMapperNames() {
        // TODO : do not use a system property for this ???
        String mns = System.getProperty("JONAS_CMP_MAPPERNAMES");
        mapperList = new ArrayList();
        if (mns != null && mns.length() > 0) {
            StringTokenizer st = new StringTokenizer(mns, ",", false);
            while (st.hasMoreElements()) {
                String mapper = st.nextToken().trim();
                if (mapper.startsWith(RDB_PREFIX + ".")) {
                    mapperList.add(mapper.substring((RDB_PREFIX + ".").length()));
                } else if (RDB_PREFIX.equals(mapper)) {
                    mapperList.add("generic");
                }
            }
        }
    }

    /**
     * Generates the java sources
     * @throws GenICException in error case
     */
    public void generate() throws GenICException {

        // Generates the java sources
        for (Iterator it = sources.iterator(); it.hasNext();) {
            Source src = (Source) it.next();
            src.generate();
        }

        // In case of Cmp2, call jorm compiler.
        // The list of mappers is given to the Jorm Compiler Configurator.
        if (dd instanceof EntityCmp2Desc) {
            EntityCmp2Desc ecd = (EntityCmp2Desc) dd;

            // Configure Jorm Compiler
            JormCompiler jc = new JormCompiler();
            JormCompilerConfigurator jcc = jc.getCompilerConfigurator();
            Properties prop = new Properties();
            prop.put("jorm.generator", "org.objectweb.jorm.generator.lib.JormGenerator");
            prop.put("jorm.mimanager", "org.objectweb.jorm.metainfo.lib.JormManager");
            prop.put("jorm.writer", "org.objectweb.jorm.mi2xml.lib.BasicDomWriter");
            prop.put("jorm.mapper.list", "rdb");
            prop.put("jorm.mapper.mifactory.rdb", "org.objectweb.jorm.mapper.rdb.metainfo.RdbMappingFactory");
            prop.put("jorm.mapper.mopfactory.rdb", "org.objectweb.jorm.mapper.rdb.generator.RdbMOPFactory");
            prop.put("jorm.mapper.gcmapping.rdb", "org.objectweb.jorm.mapper.rdb.genclass.RdbGenClassMapping");
            prop.put("jorm.mapper.schmgr.rdb", "org.objectweb.jorm.mapper.rdb.lib.RdbPMappingStructuresManager");
            prop.put("jorm.mapper.writer.rdb", "org.objectweb.jorm.mapper.rdb.mi2xml.RdbDomtreeBuilder");
            prop.put("use.context.classloader", "true");
            StringBuffer ml = new StringBuffer();
            if (!mapperList.isEmpty()) {
                Iterator it = mapperList.iterator();
                ml = ml.append((String) it.next());
                while (it.hasNext()) {
                    ml = ml.append("," + (String) it.next());
                }
            }
            prop.put("jorm.mapper.submappers.rdb", ml.toString());
            jcc.configure(prop);
            jcc.setLoggerFactory(new LoggerImpl());

            // Run the Jorm Compiler
            JormCompilerParameter cp = jc.getCompilerParameter();
            //cp.setLoggerFactory(new LoggerImpl());
            //cp.set)logger = Log.getLogger(Log.JONAS_GENIC_JORMC_PREFIX);
            cp.setProjectName(RdbMappingBuilder.getProjectName());
            cp.setKeepSrc(true);
            cp.setOutput(outdir);
            switch (ecd.getLockPolicy()) {
                case EntityDesc.LOCK_CONTAINER_READ_UNCOMMITTED:
                    cp.setBindingInheritance("org.objectweb.jonas_ejb.container.JEntitySwitchCRU");
                    break;
                case EntityDesc.LOCK_CONTAINER_SERIALIZED:
                    cp.setBindingInheritance("org.objectweb.jonas_ejb.container.JEntitySwitchCS");
                    break;
                case EntityDesc.LOCK_CONTAINER_READ_COMMITTED:
                    cp.setBindingInheritance("org.objectweb.jonas_ejb.container.JEntitySwitchCRC");
                    break;
                case EntityDesc.LOCK_DATABASE:
                    cp.setBindingInheritance("org.objectweb.jonas_ejb.container.JEntitySwitchDB");
                    break;
                default:
                    throw new GenICException("Cannot find JEntitySwitch: Unknown lock policy");
            }
            cp.setClassMappingInheritance("org.objectweb.jonas_ejb.container.jorm.RdbFactory");

            Collection jormflist = null;
            try {
                jc.setMIManager(manager);
                jormflist = jc.generateFiles(ecd.getJormList());
                // For debug only: Generate the '.pd' Jorm Files
                if (Log.getLogger(Log.JONAS_GENIC_JORMC_PREFIX).isLoggable(BasicLevel.DEBUG)) {
                    jc.generateJormFiles(ecd.getJormList());
                }
            } catch (Exception e) {
                throw new GenICException("Problem during jorm generation", e);
            }
            // Save the list of Jorm generated files
            for (Iterator i = jormflist.iterator(); i.hasNext();) {
                String file = (String) i.next();
                logger.log(BasicLevel.DEBUG, "Jorm generated file: " + file);
                noRemoteJava.add(file);
            }
        }
    }

    /**
     * @return Return the bean's name
     */
    public String getEjbName() {
        return dd.getEjbName();
    }

    /**
     * @return Return the file name of the generated source for the Home (null if none)
     */
    public String getWrpHomeFileName() {
        return wrpHomeFileName;
    }

    /**
     * @return Return the file name of the cluster configuration for the Home (null if
     * none)
     */
    public String getWrpHomeClusterConfFileName() {
        return wrpHomeClusterConfFileName;
    }

    /**
     * @return Return the file name of the generated source for the Remote (null if none)
     */
    public String getWrpRemoteFileName() {
        return wrpRemoteFileName;
    }

    /**
     * @return Return the class name of the generated source for the Remote (package included)
     */
    public String getWrpRemoteClassName() {
        return dd.getFullWrpRemoteName();
    }

    /**
     * @return Return the file name of the generated source for the ServiceEndpoint (null if none)
     */
    public String getWrpServiceEndpointFileName() {
        return wrpServiceEndpointFileName;
    }

    /**
     * @return Return the class name of the generated source for the ServiceEndpoint (package included)
     */
    public String getWrpServiceEndpointClassName() {
        if (dd instanceof SessionStatelessDesc) {
            return ((SessionStatelessDesc) dd).getFullWrpServiceEndpointName();
        }
        return null;
    }

    /**
     * @return Return the class name of the generated source for the Home (package included)
     */
    public String getWrpHomeClassName() {
        return dd.getFullWrpHomeName();
    }

    /**
     * @return Returns the list of the sources that are not Remote
     */
    public Collection getNoRemoteJavas() {
        return noRemoteJava;
    }
}
